"""
Abstract Syntax Tree (AST) node classes. The AST closely resembles
the syntax of File Selection Language, so it's helpful to keep the
language specification at hand.

The AST classes are generally struct-like and don't validate
their input values. 
"""

import os

import globals
import lexer
import utils

_running_id_number = 1

_EXPR_TYPE_DICT = {
    globals.E_BOOLEAN: 'boolean',
    globals.E_INT: 'int',
    globals.E_FLOAT: 'float',
    globals.E_STRING: 'string',
    globals.E_FILENAME: 'filename',
    globals.E_DATETIME: 'datetime',

    globals.E_STRING_FILENAME_DATETIME: '<string or filename or datetime>',
}

def type2str(expr_type):
    """Return description string for expression type, which
    is one of the constants globals.E_*. Raise ValueError
    for invalid expr_type. None is accepted as input, however,
    and means an unknown type."""
    
    if expr_type is None:
        return '<UNKNOWN>'
    if expr_type not in _EXPR_TYPE_DICT:
        raise ValueError, 'Unknown expression type: %s' % expr_type
    return _EXPR_TYPE_DICT[expr_type]


class ASTNode:
    """
    Abstract base class for AST nodes. An ASTNode instance
    knows its parent node (which may be None for the root node)
    and the root directory that is used for glob patterns.
    
    ASTNode implements the Visitor pattern with the visit method.
    Visitor object are instances of the Visitor.Visitor class.
    """
    
    def __init__(self, parent, rootdir):
        if not rootdir:
            raise ValueError, 'Rootdir not set'

        assert '//' not in rootdir
        assert '\\' not in rootdir
        self.parent = parent
        self.rootdir = utils.abspath(rootdir)
        self.token = None

        global _running_id_number
        self.idnumber = _running_id_number
        _running_id_number += 1

    def visit(self, visitor, arg=None):
        """Call the visit_XXX method corresponding to this
        class on the visitor object."""
        class_name = self.__class__.__name__
        method = getattr(visitor, 'visit_' + class_name)
        return method(self, arg)

    def set_rootdir(self, rootdir):
        self.rootdir = utils.abspath(rootdir)

    def get_parents(self):
        """Return list of parent nodes in order from direct parent
        to root node."""
        parents = []
        cur_parent = self.parent
        while cur_parent:
            parents.append(cur_parent)
            cur_parent = cur_parent.parent
        return parents

    def under_in_block(self):
        """Return whether node is an IN-block or a (grand)child of an IN-block."""
        for parent in self.get_parents():
            if parent.rootdir != self.rootdir:
                return True
        return False

### Rules ##############################################################

class Rule(ASTNode):
    """Abstract base class for rule nodes."""
    
    def __init__(self, parent, rootdir):
        ASTNode.__init__(self, parent, rootdir)

    def innerdir_common_prefix(self):
        """Return the common prefix of all innerdirs of the children
        of the node. Innerdir is a property of glob patterns; see
        documentation of GlobPattern for details."""
        raise NotImplementedError

class BlockRule(Rule):
    """
    Abstract base class for block rules: rules that contain other
    rules. The list of rules are stored in attribute 'rules'.
    """
    
    def __init__(self, parent, rootdir):
        Rule.__init__(self, parent, rootdir)
        self.rules = []

    def __str__(self):
        if self.parent:
            parentstr = '%d' % self.parent.idnumber
        else:
            parentstr= '-'
        return 'Block rule (length %d, id %s, parent %s, rootdir "%s")' % (len(self.rules),
            self.idnumber, parentstr, self.rootdir)

    def innerdir_common_prefix(self):
        children_prefixes = [c.innerdir_common_prefix() for c in self.rules]
        return os.path.commonprefix(children_prefixes)


class IfBlockRule(BlockRule):
    """
    IF-block is a rule block guarded by an expression (if-expression).
    The if-expression is stored in attribute 'if_expr' and is an
    Expression instance.
    """
    
    def __init__(self, parent, rootdir):
        BlockRule.__init__(self, parent, rootdir)
        self.if_expr = None

    def __str__(self):
        if self.parent:
            parentstr = '%d' % self.parent.idnumber
        else:
            parentstr= '-'
        return 'IF-block rule (length %d, id %s, parent %s, rootdir "%s")' % (len(self.rules),
            self.idnumber, parentstr, self.rootdir)


class SelectorRule(Rule):
    """
    Abstract base class for atomic (non-block) rules. These
    rules select files for inclusion or exclusion. The boolean
    attribute 'inclusive' marks whether a rule instance is inclusive
    or exclusive.
    """
    
    def __init__(self, parent, rootdir, inclusive):
        Rule.__init__(self, parent, rootdir)
        self.inclusive = inclusive


class GloblistRule(SelectorRule):
    """
    Glob-list is a list of glob rules. The attribute
    'glob_patterns' stores the patterns; they are instances
    of GlobPattern.
    
    The glob-list rule may have an optional if-expression.
    It is stored as an Expression instance in the 'if_expr'
    attribute and it is None if there is no if-expression
    associated with the glob-list rule. 
    """
    
    def __init__(self, parent, rootdir, inclusive):
        SelectorRule.__init__(self, parent, rootdir, inclusive)
        self.glob_patterns = []
        self.if_expr = None

    def __str__(self):
        if self.parent:
            parentstr = '%d' % self.parent.idnumber
        else:
            parentstr= '-'
        if self.inclusive:
            exclstr = ''
        else:
            exclstr = 'EXCLUDE '
        return 'Glob list %s(length %d, parent %s, rootdir "%s")' % (exclstr, 
            len(self.glob_patterns), parentstr, self.rootdir)

    def innerdir_common_prefix(self):
        children_prefixes = [c.innerdir for c in self.glob_patterns]
        return os.path.commonprefix(children_prefixes)


class ForEachRule(SelectorRule):
    """
    For-each rule contains a list of glob patterns and an
    expression that is evaluated for each file (in the interpreter).
    The glob patterns are stored in attribute 'glob_patterns' and
    the if-expression in 'if_expr'. There is also a variable expression,
    in attribute 'variable_expr', that is an instance of VariableExpression.
    """
    
    def __init__(self, parent, rootdir, inclusive):
        SelectorRule.__init__(self, parent, rootdir, inclusive)
        self.variable_expr = None
        self.glob_patterns = []
        self.if_expr = None

    def __str__(self):
        if self.inclusive:
            exclstr = ''
        else:
            exclstr = 'EXCLUDE '
        return 'For each %s(variable "%s", %s globs, rootdir "%s")' % (exclstr,
            self.variable_expr.variable_name, len(self.glob_patterns), self.rootdir)

    def innerdir_common_prefix(self):
        children_prefixes = [c.innerdir for c in self.glob_patterns]
        return os.path.commonprefix(children_prefixes)


class GlobPattern(ASTNode):
    """
    A single glob pattern, the most basic rule-level unit.
    
    Attributes:
      - 'pattern' is the glob pattern as read from source file. It is
         normalized by replacing backslashes with forward slashes.
      - 'recursive' tells whether the pattern should be evaluated
         in subdirectories recursively.
      - 'is_absolute' tells whether the pattern is relative (to rootdir)
        or absolute.
      - 'fullglob' is an absolute version of the pattern, i.e. it begins
        with a full directory specification. Like 'pattern', 'fullglob'
        is normalized to forward slases.
      - 'innerdir' is the innermost directory where the pattern should
        be evaluated. In other word, the evaluation start here. Like
        'fullglob', 'innerdir' is an absolute file name and normalized.
    """
    
    def __init__(self, parent, rootdir, pattern, recursive):
        ASTNode.__init__(self, parent, rootdir)
        self.pattern = pattern.replace('\\', '/')
        self.recursive = recursive
        self.is_absolute = utils.filename_is_absolute(self.pattern)

        if not self.is_absolute:
            self.fullglob, self.innerdir = utils.calc_fullglob_innerdir(self.rootdir, self.pattern)
        else:
            self.fullglob = utils.abspath(self.pattern)
            self.innerdir = utils.calc_innerdir(self.fullglob)

    def __str__(self):
        if not self.recursive:
            return 'Glob: "%s" NONREC (innerdir "%s")' % (self.fullglob, self.innerdir)
        else:
            return 'Glob: "%s" (innerdir "%s")' % (self.fullglob, self.innerdir)

### Expressions ########################################################

class Expression(ASTNode):
    """
    Abstract base class for expressions. Expression type
    (a constant globals.E_*) is stored in attribute 'expr_type'.
    """
    
    def __init__(self, parent, rootdir, expr_type):
        ASTNode.__init__(self, parent, rootdir)
        self.expr_type = expr_type


class AndExpression(Expression):
    """
    Boolean AND expression type. The expressions are stored in
    attributes 'expr1' and 'expr2' and are instances of Expression.
    """
    
    def __init__(self, parent, rootdir):
        Expression.__init__(self, parent, rootdir, globals.E_BOOLEAN)
        self.expr1 = None
        self.expr2 = None

    def __str__(self):
        return 'AND expression'


class OrExpression(Expression):
    """
    Boolean OR expression type. The expressions are stored in
    attributes 'expr1' and 'expr2' and are instances of Expression.
    """
    
    def __init__(self, parent, rootdir):
        Expression.__init__(self, parent, rootdir, globals.E_BOOLEAN)
        self.expr1 = None
        self.expr2 = None

    def __str__(self):
        return 'OR expression'


class NotExpression(Expression):
    """
    Boolean NOT expression type. The expression to be negated is stored in
    attribute 'expr' and is an instance of Expression.
    """
    
    def __init__(self, parent, rootdir,):
        Expression.__init__(self, parent, rootdir, globals.E_BOOLEAN)
        self.expr = None

    def __str__(self):
        return 'NOT expression'


class FunctionExpression(Expression):
    """
    Function call expression. The name of the function is stored
    as a string in 'funcname' and the paremeters in 'params'. The
    parameters are instances of Expression. 
    """
    
    def __init__(self, parent, rootdir, funcname):
        Expression.__init__(self, parent, rootdir, None)
        self.funcname = funcname
        self.params = [] # [Expression]

    def __str__(self):
#        paramstr = ', '.join([str(p) for p in self.params])
        return 'Function call (funcname "%s", %d params)' % (self.funcname, len(self.params))


class CompareExpression(Expression):
    """
    Comparison operator expression. The expressions to be compared
    are stored in attributes 'expr1' and 'expr2' and are instances of
    Expression. The comparison operator is in 'operator_type' and
    is a token type defined in lexer.T_*. 
    """
    
    def __init__(self, parent, rootdir, expr1, operator_type):
        Expression.__init__(self, parent, rootdir, globals.E_BOOLEAN)
        self.expr1 = expr1
        self.expr2 = None
        self.operator_type = operator_type

    def __str__(self):
        s = lexer.type2str(self.operator_type)
        return 'Compare expression (type "%s")' % s


class LiteralExpression(Expression):
    """
    Number, string, filename or datetime literal. The string value
    is stored in 'strvalue'. For datetime literals, the datetime value
    is stored as an datetime.datetime instance in 'datetime_value'.
    For filename literals, the value is stored in 'filename_value'. 
    """
    
    def __init__(self, parent, rootdir, strvalue, atom_type):
        """
        Init.
        'atom_type' is an globals.E_* type that marks the type
        of the literal.
        """
        
        Expression.__init__(self, parent, rootdir, atom_type)
        self.strvalue = strvalue
        self.datetime_value = None
        self.filename_value = None

    def __str__(self):
        if self.expr_type in (globals.E_STRING, globals.E_STRING_FILENAME_DATETIME):
            s = '"%s"' % self.strvalue
        elif self.expr_type == globals.E_FILENAME:
            s = self.filename_value
        elif self.expr_type == globals.E_DATETIME:
            s = self.datetime_value.strftime('%Y-%m-%d %H:%M:%S')
        else:
            s = self.strvalue
        return 'Atom (%s, type %s)' % (s, type2str(self.expr_type))


class VariableExpression(Expression):
    """
    Reference to a variable, as used in an expression. The variable
    name in stored in 'variable_name' and a reference to an
    symboltable.Variable instance is stored in 'variable_obj.
    The value of the variable is accessed via 'variable_obj'.
    """
    
    def __init__(self, parent, rootdir, variable_name):
        Expression.__init__(self, parent, rootdir, None)
        self.variable_name = variable_name
        self.variable_obj = None # Reference to symboltable.Variable

    def __str__(self):
        return 'Variable (%s, type %s)' % (self.variable_name, type2str(self.expr_type))
