"""
Contains the abstract base class for hook handlers
that real handlers expand.

Hook handlers receive real-time information on
interpretation and file-system scanning. Inclusion
and exclusion callback functions also receive the
source rule that was applied.

The hook mechanism is one-direction only; hook
handlers can't modify the interpreter during running,
only receive information on the proceeding.
"""

class HookHandler:
    """Base class for hook handlers. The default implementations
    for the methods do nothing."""

    def include_file(self, fileobj, rule):
        """Called when a file has been included in the
        file set. Only called once for each file.

        :param fileobj: The included file as `File.File`
          instance.
        :param rule: The rule that resulted in the inclusion.
          `AST.Rule` instance.
        """
        pass

    def exclude_file(self, fileobj, rule):
        """Called when a file has been included in the
        file set (NOTE: CURRENTLY NEVER CALLED). Only called once
        for each file.

        :param fileobj: The excluded file as `File.File`
          instance.
        :param rule: The rule that resulted in the exclusion.
          `AST.Rule` instance.
        """
        pass

    def start_scan(self, dirnameobj, glob_pattern):
        """Called when a directory scan starts.

        :param dirnameobj: Scanned directory as `File.File`
          instance.
        :param glob_pattern: The glob pattern that is used
          in scanning, as string.
        """
        pass

    def end_scan(self, dirnameobj, files_count, bytes_count):
        """Called when a directory scan finishes.

        :param dirnameobj: Scanned directory as `File.File`
          instance.
        :param files_count: Number of files that were included
          in the directory, including its sub-directories.
        :param bytes_count: Size in bytes of the included files
          in the directory, including its sub-directories.
        """
        pass
