"""
Visitor for AST type inference and type checking.
"""

import re

import AST
import fslparser
import functions
import globals
import symboltable
import Visitor
import utils

def compare_compatible(type1, type2):
    """Return whether values of types 'type1' and 'type2' can
    be compared."""
    return ( ((type1 == type2) and (type1 != globals.E_BOOLEAN))
         or type1 == globals.E_INT and type2 == globals.E_FLOAT
         or type1 == globals.E_FLOAT and type2 == globals.E_INT
         or (type1 == globals.E_STRING_FILENAME_DATETIME and type2 in (globals.E_STRING, globals.E_FILENAME, globals.E_DATETIME))
         or (type2 == globals.E_STRING_FILENAME_DATETIME and type1 in (globals.E_STRING, globals.E_FILENAME, globals.E_DATETIME)) )

def assignment_compatible(type1, type2):
    """Return whether value of type1 can assigned into a variable
    of type2. In other words, the operation variable2 := value1 is legal."""
    return (type1 == type2
         or type1 == globals.E_INT and type2 == globals.E_FLOAT
         or type1 == globals.E_FLOAT and type2 == globals.E_INT
         or (type1 == globals.E_STRING_FILENAME_DATETIME and type2 in (globals.E_STRING, globals.E_FILENAME, globals.E_DATETIME))
         or (type2 == globals.E_STRING_FILENAME_DATETIME and type1 in (globals.E_STRING, globals.E_FILENAME, globals.E_DATETIME)) )

class TypeChecker(Visitor.Visitor):
    """
    Infer and check expression types and decorate AST with type info.
    Also construct symbol table and set the variable_obj field
    in VariableExpression instances. (The symbol table is not returned
    itself.)
    """

    def __init__(self):
        self.symboltable = symboltable.SymbolTable()

    def _check_type_set(self, astnode):
        if astnode.expr_type is None:
            raise globals.FSLSyntaxError(astnode.token, 'Internal Error: expression type is unknown (%s)' % astnode)

    def _check_type_boolean(self, astnode):
        if astnode.expr_type != globals.E_BOOLEAN:
            msg = 'Type of expression must be boolean (is: %s)' % AST.type2str(astnode.if_expr.expr_type)
            raise globals.FSLSyntaxError(astnode.token, msg)

    def visit_BlockRule(self, astnode, arg):
        for rule in astnode.rules:
            rule.visit(self)

    def visit_IfBlockRule(self, astnode, arg):
        astnode.if_expr.visit(self)
        for rule in astnode.rules:
            rule.visit(self)

    def visit_GloblistRule(self, astnode, arg):
        if astnode.if_expr:
            astnode.if_expr.visit(self)
            if astnode.if_expr.expr_type != globals.E_BOOLEAN:
                msg = 'Type of IF-expression must be boolean (is: %s)' % AST.type2str(astnode.if_expr.expr_type)
                raise globals.FSLSyntaxError(astnode.token, msg)

    def visit_ForEachRule(self, astnode, arg):
#        astnode.variable_expr.expr_type = globals.E_FILENAME
        varobj = symboltable.Variable(astnode.variable_expr.variable_name, globals.E_FILENAME)
        astnode.variable_expr.expr_type = varobj.variable_type
        astnode.variable_expr.variable_obj = varobj
        self.symboltable.push(varobj)
        astnode.if_expr.visit(self)
        self.symboltable.pop()
        if astnode.if_expr.expr_type != globals.E_BOOLEAN:
            msg = 'Type of IF-expression must be boolean (is: %s)' % AST.type2str(astnode.if_expr.expr_type)
            raise globals.FSLSyntaxError(astnode.token, msg)

    def visit_AndExpression(self, astnode, arg):
        astnode.expr1.visit(self)
        astnode.expr2.visit(self)
        self._check_type_boolean(astnode.expr1)
        self._check_type_boolean(astnode.expr2)

    def visit_OrExpression(self, astnode, arg):
        astnode.expr1.visit(self)
        astnode.expr2.visit(self)
        self._check_type_boolean(astnode.expr1)
        self._check_type_boolean(astnode.expr2)

    def visit_NotExpression(self, astnode, arg):
        astnode.expr.visit(self)
        self._check_type_boolean(astnode.expr)

    def visit_FunctionExpression(self, astnode, arg):
        return_type = functions.get_function_return_type(astnode.funcname)
        param_types = functions.get_function_param_types(astnode.funcname)

        if len(param_types) != len(astnode.params):
            msg = 'Function "%s" takes %d arguments, %d given' % (astnode.funcname, len(param_types), len(astnode.params))
            raise globals.FSLSyntaxError(astnode.token, msg)

        for param_expr, param_type in zip(astnode.params, param_types):
            param_expr.visit(self)
            if (param_expr.expr_type is None) or (param_expr.expr_type == globals.E_STRING_FILENAME_DATETIME):
                param_expr.expr_type = param_type
                param_expr.visit(self)
            elif param_expr.expr_type != param_type:
                msg = 'Invalid argument for function "%s": expected type "%s", got "%s"' % (
                    astnode.funcname, AST.type2str(param_type), AST.type2str(param_expr.expr_type))
                raise globals.FSLSyntaxError(astnode.token, msg)

        astnode.expr_type = return_type

    def visit_CompareExpression(self, astnode, arg):
        astnode.expr1.visit(self)
        astnode.expr2.visit(self)

        type1 = astnode.expr1.expr_type
        type2 = astnode.expr2.expr_type

        if (type1 is None) and (type2 is not None) or (type1==globals.E_STRING_FILENAME_DATETIME and type2 not in (None, globals.E_STRING_FILENAME_DATETIME)):
            if assignment_compatible(type1, type2):
                astnode.expr1.expr_type = type2
                astnode.expr1.visit(self) # Must visit again
                type1 = astnode.expr1.expr_type
            else:
                msg = "Can't compare types '%s' and '%s'" % (AST.type2str(type1), AST.type2str(type2))
                raise globals.FSLSyntaxError(astnode.token, msg)
        elif (type2 is None) and (type1 is not None) or (type2==globals.E_STRING_FILENAME_DATETIME and type1 not in (None, globals.E_STRING_FILENAME_DATETIME)):
            if assignment_compatible(type2, type1):
                astnode.expr2.expr_type = type1
                astnode.expr2.visit(self) # Must visit again
                type2 = astnode.expr2.expr_type
            else:
                msg = "Can't compare types '%s' and '%s'" % (AST.type2str(type1), AST.type2str(type2))
                raise globals.FSLSyntaxError(astnode.token, msg)

        self._check_type_set(astnode.expr1)
        self._check_type_set(astnode.expr2)

        if not compare_compatible(type1, type2):
            msg = "Can't compare types '%s' and '%s'" % (AST.type2str(type1), AST.type2str(type2))
            raise globals.FSLSyntaxError(astnode.token, msg)

    def visit_LiteralExpression(self, astnode, arg):
        if astnode.expr_type == globals.E_DATETIME:
            try:
                datetime_value = utils.parse_datetime(astnode.strvalue)
            except ValueError:
                msg = 'Invalid datetime literal: "%s" (supported format: ISO)' % astnode.strvalue
                raise globals.FSLSyntaxError(astnode.token, msg)
            astnode.datetime_value = datetime_value
        elif astnode.expr_type == globals.E_FILENAME:
            if utils.filename_is_absolute(astnode.strvalue):
                astnode.filename_value = astnode.strvalue
            else:
                astnode.filename_value = utils.abspath(utils.combine_rootdir(astnode.rootdir, astnode.strvalue))

    def visit_VariableExpression(self, astnode, arg):
        varobj = self.symboltable.get(astnode.variable_name)

        if varobj is None:
            msg = 'Unknown variable: "%s"' % astnode.variable_name
            raise globals.FSLSyntaxError(astnode.token, msg)

        astnode.variable_obj = varobj
        astnode.expr_type = varobj.variable_type
