"""
Parser reads FSL source files and produces an AST from
them.
"""

import cStringIO

import functions
import AST
import ASTChecker
import ASTLinearizer
import ASTPrinter
import globals
import lexer
import TypeChecker
import utils

_DEBUG = False

class _Parser:
    # This class is not usually used directly.
    
    def __init__(self, ast, filename, rootdir):
        self.ast = ast   # BlockRule instance
        self.filename = filename
        self.rootdir = rootdir.replace('\\', '/')

        self._curtoken = None
        self._lexer = lexer.iter_tokens(filename)

    def _read_token(self):
        try:
            self._curtoken = self._lexer.next()
        except StopIteration:
            self._curtoken = None
    
    def _match_read(self, tokentypes):
        self._check_type_force(tokentypes)
        if _DEBUG:
            print 'Consumed %s' % self._curtoken
        self._read_token()

    def _check_type(self, tokentypes):
        if type(tokentypes) == int:
            tokentypes = [tokentypes]
        return (self._curtoken) and (self._curtoken.tokentype in tokentypes)

    def _check_type_force(self, tokentypes):
        if type(tokentypes) == int:
            tokentypes = [tokentypes]
        if not self._check_type(tokentypes):
            if len(tokentypes) == 1:
                msg = 'Expected %s' % lexer.type2str(tokentypes[0])
            else:
                types = ', '.join([lexer.type2str(t) for t in tokentypes])
                msg = 'Expected one of %s' % types
            raise globals.FSLParseError, (self._curtoken, msg)

    def parsefile(self):
        self._read_token()
        blockrule = self._parse_blockrule(self.ast, self.rootdir)
        assert blockrule
        assert blockrule.rules
        self._match_read(lexer.T_EOF)
        self.ast.rules.append(blockrule)

    def parsestring(self, string):
        strfile = cStringIO.StringIO(string)
        self._lexer = lexer.iter_tokens_fileobj(strfile, self.filename)
        self._read_token()
        blockrule = self._parse_blockrule(self.ast, self.rootdir)
        assert blockrule
        assert blockrule.rules
        self._match_read(lexer.T_EOF)
        self.ast.rules.append(blockrule)
        strfile.close()

    ### Parser methods #################################################

    # Some helpers.

    def _parse_rule_sequence(self, parent, rootdir):
        """Parse rule sequence, return list of Rule instance."""
        # (<rule> LINE-END)+
        rules = []
        first = (lexer.T_STRING, lexer.T_BARE_STRING, lexer.T_EACH, lexer.T_NOT, lexer.T_IN, lexer.T_IF)
        self._check_type_force(first)
        while self._check_type(first):
            rule = self._parse_rule(parent, rootdir)
            rules.append(rule)
        return rules

    def _parse_globlist(self, parent, rootdir):
        """Parse glob sequence, return list of GlobPattern."""
        # <glob> ("," <glob>)*
        patterns = []
        glob_pattern = self._parse_glob(parent, rootdir)
        assert glob_pattern
        patterns.append(glob_pattern)
        while self._check_type(lexer.T_COMMA):
            self._match_read(lexer.T_COMMA)
            glob_pattern = self._parse_glob(parent, rootdir)
            patterns.append(glob_pattern)
        return patterns

    def _parse_arglist(self, parent, rootdir):
        # <argument-list> := ( <atom> ("," <argument-list>)? )?
        #  FIRST = {STRING, INT, FLOAT, BARE-STRING}
        first = (lexer.T_STRING, lexer.T_INT, lexer.T_FLOAT, lexer.T_BARE_STRING)

        atoms = []
        while self._check_type(first):
            atom_expr = self._parse_atom(parent, rootdir)
            atoms.append(atom_expr)
            if not self._check_type(lexer.T_COMMA):
                break
            else:
                self._match_read(lexer.T_COMMA)

        return atoms

    # Now, the actual parse methods.

    def _parse_blockrule(self, parent, rootdir):
        # <rule>+
        #  FIRST = {STRING, BARE-STRING, EACH, NOT, IN}

        blockrule = AST.BlockRule(parent, rootdir)
        first = (lexer.T_STRING, lexer.T_BARE_STRING, lexer.T_EACH, lexer.T_NOT, lexer.T_IN, lexer.T_IF)
        self._check_type_force(first)
        while self._check_type(first):
            rule = self._parse_rule(blockrule, rootdir)
            blockrule.rules.append(rule)
        assert blockrule.rules
        return blockrule

    def _parse_rule(self, parent, rootdir):
        # <rule> := <glob-list>
        #   FIRST = {STRING, BARE-STRING}
        # <rule> := <for-each>
        #   FIRST = {EACH}
        # <rule> := NOT (<glob-list> | <for-each>)
        #   FIRST = {NOT}
        # <rule> := IN BARE-STRING <start-block> <rule>+ <end-block>
        #   FIRST = {IN}
        # <rule> := IF <expression> <start-block> <rule>+ <end-block>
        #   FIRST = {IF}

        if self._check_type(lexer.T_NOT):
            self._match_read(lexer.T_NOT)
            inclusive = False
        else:
            inclusive = True

        if self._check_type((lexer.T_STRING, lexer.T_BARE_STRING)):
            rule = self._parse_globlist_rule(parent, rootdir, inclusive)
            assert rule.glob_patterns
        elif self._check_type(lexer.T_EACH):
            rule = self._parse_each(parent, rootdir, inclusive)
        elif self._check_type(lexer.T_IN):
            if not inclusive:
                # NOT was given for an IN-block
                raise globals.FSLParseError(self._curtoken, "NOT is not allowed for an IN-block")
            rule = AST.BlockRule(parent, rootdir)
            new_rootdir, rule.rules = self._parse_in_block(rule, rootdir)
            rule.set_rootdir(new_rootdir)
            assert rule.rules
        elif self._check_type(lexer.T_IF):
            if not inclusive:
                # NOT was given for an IF-block
                raise globals.FSLParseError(self._curtoken, "NOT is not allowed for an IF-block")
            rule = self._parse_if_block(parent, rootdir)
            assert rule.rules
        else:
            raise globals.FSLParseError(self._curtoken, 'Expected glob list, EACH-rule or IN-block')

        assert rule
        return rule

    def _parse_globlist_rule(self, parent, rootdir, inclusive):
        # <rule> := (NOT)? <glob-list> (IF <expression>)?
        #  FIRST = {NOT, STRING, BARE-STRING}
        # Note: NOT was consumed earlier.
        glrule = AST.GloblistRule(parent, rootdir, inclusive)
        glrule.glob_patterns = self._parse_globlist(glrule, rootdir)

        if self._check_type(lexer.T_IF):
            self._match_read(lexer.T_IF)
            glrule.if_expr = self._parse_expression(glrule, rootdir)

        self._match_read(lexer.T_LINE_END)
        return glrule

    def _parse_glob(self, parent, rootdir):
        # <glob> := (STRING | BARE-STRING) (NONREC)?
        #   FIRST = {STRING, BARE-STRING}

        first = (lexer.T_STRING, lexer.T_BARE_STRING)
        self._check_type_force(first)
        globstr = self._curtoken.strvalue

        if utils.filename_is_absolute(globstr) and parent is not None and parent.under_in_block():
            raise globals.FSLParseError(self._curtoken, 'Glob pattern inside IN-block must be relative')

        self._match_read(first)

        if self._check_type(lexer.T_NONREC):
            self._match_read(lexer.T_NONREC)
            recursive = False
        else:
            recursive = True

        glob_pattern = AST.GlobPattern(parent, rootdir, globstr, recursive)
        return glob_pattern

    def _parse_each(self, parent, rootdir, inclusive):
        # <for-each> := EACH BARE-STRING (IN <glob-list>)? IF <expression>
        #   FIRST = {EACH}

        self._match_read(lexer.T_EACH)

        self._check_type_force(lexer.T_BARE_STRING)
        variable_name = self._curtoken.strvalue
        self._match_read(lexer.T_BARE_STRING)

        this_rule = AST.ForEachRule(parent, rootdir, inclusive)
        this_rule.variable_expr = AST.VariableExpression(this_rule, rootdir, variable_name)

        if self._check_type(lexer.T_IN):
            self._match_read(lexer.T_IN)
            this_rule.glob_patterns = self._parse_globlist(this_rule, rootdir)
        else:
            all_glob = AST.GlobPattern(this_rule, rootdir, '*', True)
#            all_glob.calc_extra_globs()
            this_rule.glob_patterns = [all_glob]

        self._match_read(lexer.T_IF)
        this_rule.if_expr = self._parse_expression(this_rule, rootdir)

        self._match_read(lexer.T_LINE_END)

        assert this_rule.variable_expr.variable_name
        assert this_rule.glob_patterns
        return this_rule

    def _parse_in_block(self, parent, rootdir):
        # <rule> := IN (BARE-STRING|STRING) <start-block> <rule>+ <end-block>
        #  FIRST = {IN}

        self._match_read(lexer.T_IN)
        self._check_type_force((lexer.T_BARE_STRING, lexer.T_STRING))
        directory = self._curtoken.strvalue

        if utils.filename_is_absolute(directory) and parent is not None and parent.under_in_block():
            raise globals.FSLParseError(self._curtoken, 'Directory of nested IN-block must be relative')

        self._match_read((lexer.T_BARE_STRING, lexer.T_STRING))
        self._match_read(lexer.T_LINE_END)
        self._match_read(lexer.T_INDENT)

        this_rootdir = utils.combine_rootdir(rootdir, directory)
        rules = self._parse_rule_sequence(parent, this_rootdir)

        self._match_read(lexer.T_DEDENT)
        return this_rootdir, rules

    def _parse_if_block(self, parent, rootdir):
        # <rule> := IF <expression> <start-block> <rule>+ <end-block>
        #  FIRST = {IF}

        self._match_read(lexer.T_IF)
        ifrule = AST.IfBlockRule(parent, rootdir)
        ifrule.if_expr = self._parse_expression(ifrule, rootdir)
        self._match_read(lexer.T_LINE_END)
        self._match_read(lexer.T_INDENT)
        ifrule.rules = self._parse_rule_sequence(ifrule, rootdir)
        self._match_read(lexer.T_DEDENT)

        return ifrule

    def _parse_expression(self, parent, rootdir):
        # <expression> := <simple-expression> ((AND | OR) <expression>)?
        #   FIRST = {BARE-STRING, "("}

        expr = self._parse_simple_expression(parent, rootdir)

        if self._check_type(lexer.T_AND):
            self._match_read(lexer.T_AND)
            and_expr = AST.AndExpression(parent, rootdir)
            expr.parent = and_expr
            and_expr.expr1 = expr
            and_expr.expr2 = self._parse_simple_expression(and_expr, rootdir)
            return and_expr
        elif self._check_type(lexer.T_OR):
            self._match_read(lexer.T_OR)
            or_expr = AST.OrExpression(parent, rootdir)
            expr.parent = or_expr
            or_expr.expr1 = expr
            or_expr.expr2 = self._parse_simple_expression(or_expr, rootdir)
            return or_expr
        else:
            return expr

    def _parse_simple_expression(self, parent, rootdir):
        # <simple-expression> := (NOT)? <atom> (<rel-op> <atom>)?
        #   FIRST = {NOT, BARE-STRING}
        # <simple-expression> := (NOT)? "(" <expression> ")"
        #   FIRST = {NOT, "("}

        if self._check_type(lexer.T_NOT):
            self._match_read(lexer.T_NOT)
            not_expr = AST.NotExpression(parent, rootdir)
            not_expr.expr = self._parse_simple_expression(not_expr, rootdir)
            return not_expr

        if self._check_type(lexer.T_LEFT_PAREN):
            self._match_read(lexer.T_LEFT_PAREN)
            expr = self._parse_expression(parent, rootdir)
            self._match_read(lexer.T_RIGHT_PAREN)
            return expr

        """
        self._check_type_force(lexer.T_BARE_STRING)
        funcname = self._curtoken.strvalue
        functype = functions.get_function_type(funcname)
        func_expr = AST.FunctionExpression(parent, rootdir, funcname, functype)
        self._match_read(lexer.T_BARE_STRING)
        self._match_read(lexer.T_LEFT_PAREN)
        param_expr = self._parse_atom(func_expr, rootdir)
        func_expr.params = [param_expr]
        self._match_read(lexer.T_RIGHT_PAREN)
        """
        atom_expr = self._parse_atom(parent, rootdir)

        relfirst = lexer.T_LESS_THAN, lexer.T_LESS_THAN_EQUAL, lexer.T_MORE_THAN, lexer.T_MORE_THAN_EQUAL, lexer.T_EQUAL, lexer.T_NOT_EQUAL
        if self._check_type(relfirst):
            rel_type = self._curtoken.tokentype
            self._match_read(relfirst)
            rel_expr = AST.CompareExpression(parent, rootdir, atom_expr, rel_type)
            atom_expr.parent = rel_expr
            rel_expr.expr2 = self._parse_atom(rel_expr, rootdir)
            return rel_expr
        else:
            return atom_expr

    def _parse_atom(self, parent, rootdir):
        # <atom> := STRING
        #   FIRST = {STRING}
        # <atom> := <number>
        #   FIRST = {INT, FLOAT}
        # <atom> := BARE-STRING ("(" <atom> ")")?
        #   FIRST = {BARE-STRING}

        if self._check_type(lexer.T_BARE_STRING):
            # Variable name or function call
            strvalue = self._curtoken.strvalue
            self._match_read(lexer.T_BARE_STRING)
            if self._check_type(lexer.T_LEFT_PAREN):
                # Function call
                funcname = strvalue
                self._match_read(lexer.T_LEFT_PAREN)
#                functype = functions.get_function_type(funcname)
                func_expr = AST.FunctionExpression(parent, rootdir, funcname)
#                param_expr = self._parse_atom(func_expr, rootdir)
                func_expr.params = self._parse_arglist(func_expr, rootdir)
                self._match_read(lexer.T_RIGHT_PAREN)
                return func_expr
            else:
                # Variable
                var_expr = AST.VariableExpression(parent, rootdir, strvalue)
                return var_expr

        elif self._check_type(lexer.T_STRING):
            strvalue = self._curtoken.strvalue
            self._match_read(lexer.T_STRING)
            # We don' yet know the exact type of the literal; it can
            # be string (E_STRING), filename (E_FILENAME) or datetime
            # (E_DATETIME). This is inferred later.
            expr = AST.LiteralExpression(parent, rootdir, strvalue, globals.E_STRING_FILENAME_DATETIME)
            return expr

        elif self._check_type(lexer.T_INT):
            strvalue = self._curtoken.strvalue
            self._match_read(lexer.T_INT)
            expr = AST.LiteralExpression(parent, rootdir, strvalue, globals.E_INT)
            return expr

        elif self._check_type(lexer.T_FLOAT):
            strvalue = self._curtoken.strvalue
            self._match_read(lexer.T_FLOAT)
            expr = AST.LiteralExpression(parent, rootdir, strvalue, globals.E_FLOAT)
            return expr

        else:
            raise globals.FSLParseError(self._curtoken, "Expected atom (STRING, INT, FLOAT or simple expression)")

def parse(filenames, rootdir):
    """
    Parse each source file in 'filenames' and return
    an AST.BlockRule instance. All the source files are combined
    into the block-rule that is returned. 'rootdir' is the root
    directory on the file system that is used for relative
    globs.
    """
    
    rootdir = rootdir.replace('\\', '/')
    ast = AST.BlockRule(None, rootdir)
    if type(filenames) in (str, unicode):
        filenames = [filenames]

    for filename in filenames:
        parser = _Parser(ast, filename, rootdir)
        parser.parsefile()

    ast.visit(TypeChecker.TypeChecker(), [])

    return ast

def parse_string(string, rootdir):
    """
    Parse a string containing FSL source and return
    an AST.BlockRule instance. 'rootdir' is the root
    directory on the file system that is used for relative
    globs.
    """
    
    rootdir = rootdir.replace('\\', '/')
    ast = AST.BlockRule(None, rootdir)

    parser = _Parser(ast, '<STRING>', rootdir)
    parser.parsestring(string)

    ast.visit(TypeChecker.TypeChecker(), [])
    return ast
