"""
Lexer reads FSL source files and divides the source into
atomic tokens.
"""

import re

import fslparser
import globals

T_INDENT = 1
T_DEDENT = 2
T_LINE_END = 3
T_EOF = 4

T_STRING = 10
T_BARE_STRING = 11
T_INT = 12
T_FLOAT = 13

T_COMMA = 20
T_LEFT_PAREN = 21
T_RIGHT_PAREN = 22
T_LESS_THAN = 23
T_LESS_THAN_EQUAL = 24
T_MORE_THAN = 25
T_MORE_THAN_EQUAL = 26
T_EQUAL = 27
T_NOT_EQUAL = 28

T_AND = 30
T_EACH = 31
T_IF = 32
T_IN = 33
T_NONREC = 34
T_NOT = 35
T_OR = 36

_T_COMMENT = 99

_TYPE_STRINGS = {
    T_INDENT: 'INDENT',
    T_DEDENT: 'DEDENT',
    T_LINE_END: 'LINE-END',
    T_EOF: 'EOF',

    T_STRING: 'STRING',
    T_BARE_STRING: 'BARE-STRING',
    T_INT: 'INT',
    T_FLOAT: 'FLOAT',

    T_COMMA: ',',
    T_LEFT_PAREN: '(',
    T_RIGHT_PAREN: ')',
    T_LESS_THAN: '<',
    T_LESS_THAN_EQUAL: '<=',
    T_MORE_THAN: '>',
    T_MORE_THAN_EQUAL: '>=',
    T_EQUAL: '=',
    T_NOT_EQUAL: '!=',

    T_AND: 'AND',
    T_EACH: 'EACH',
    T_IF: 'IF',
    T_IN: 'IN',
    T_NONREC: 'NONREC',
    T_NOT: 'NOT',
    T_OR: 'OR',
}

def type2str(t):
    """Return string describing the token type 't', which
    must be one of T_* constants."""
    return _TYPE_STRINGS.get(t, '<UNKNOWN TOKEN TYPE>')

# Possible tests:
# - no consecutive T_LINE_END tokens
# - equal number of T_INDENT and T_DEDENT tokens
# - T_EOF token at end

class Token:
    """
    Atomic syntax element in source file.
    Each token has a type (stored in attribute 'tokentype', one
    of T_* constants), string value (in 'strvalue'), source filename
    where the token comes from (in 'filename') and line number (in 'lineno').
    Numeric tokens have the numeric value in attribute 'numbervalue';
    for others, the attribute is None.
    """
    
    def __init__(self, tokentype, strvalue, filename, lineno):
        self.tokentype = tokentype
        self.strvalue = strvalue
        self.filename = filename
        self.lineno = lineno

        if self.tokentype == T_INT:
            self.numbervalue = int(strvalue, 10)
        elif self.tokentype == T_FLOAT:
            self.numbervalue = float(strvalue)
        else:
            self.numbervalue = None

    def __str__(self):
        return '"%s" (type %s, line %s)' % (self.strvalue, type2str(self.tokentype), self.lineno)

_BARE_FORBIDDEN = ' ,()"<>=()#|!'
_BARE_STRING_RE = r'[^%s0-9\n][^%s\n]*' % (_BARE_FORBIDDEN, _BARE_FORBIDDEN)

_TOKENS_RE = [
    (r'[0-9]+[.][0-9]+', T_FLOAT),
    (r'\d+', T_INT),
    (r'["]([^"\n])*["]', T_STRING),
    (_BARE_STRING_RE, T_BARE_STRING),
    (r'[,]', T_COMMA),
    (r'[(]', T_LEFT_PAREN),
    (r'[)]', T_RIGHT_PAREN),
    (r'<=', T_LESS_THAN_EQUAL),
    (r'<', T_LESS_THAN),
    (r'>=', T_MORE_THAN_EQUAL),
    (r'>', T_MORE_THAN),
    (r'=', T_EQUAL),
    (r'!=', T_NOT_EQUAL),
    (r'#', _T_COMMENT),
]

_RESERVED_WORDS = {
    'and': T_AND,
    'each': T_EACH,
    'if': T_IF,
    'in': T_IN,
    'nonrec': T_NONREC,
    'not': T_NOT,
    'or': T_OR,
}

_NEWLINE = '\n|\r|\r\n'
_WHITESPACE = ' \t'

def iter_tokens(filename):
    """Generate (yield) tokens for source file 'filename'. The
    tokens are instances of Token. Raise IOError if
    filename can't be accessed."""
    
    infile = file(filename, 'rt')
    for t in iter_tokens_fileobj(infile, filename):
        yield t
    infile.close()

def iter_tokens_fileobj(infile, filename):
    """Generate (yield) tokens for file object 'infile'.
    The tokens are instances of Token.
    'Filename' is the source filename and is used only for
    error messages."""

    line = ''
    eof = False
    lineno = 0
    prev_indent = 0
    indent_length = None
    line_had_content = False

    cur_indent = 0
    indent_count = 0
    dedent_count = 0
    open_paren_count = 0

    while not eof:
        while not line:
            line_had_content = False
            line = infile.readline()
            if not line:
                eof = True
                break
            lineno += 1
            line = line.strip('\n\r')
            strippedline = line.lstrip()
            if (not strippedline) or strippedline.startswith('#'):
                line = ''
                continue
            elif open_paren_count > 0:
                # We are inside an expression with parens, so we
                # don't yield indent or dedent tokens.
                line = strippedline
            else:
                cur_indent = len(line) - len(strippedline)
                if (indent_length is None) and cur_indent > 0:
                    indent_length = cur_indent
                if cur_indent != prev_indent:
                    if (cur_indent % indent_length) != 0:
                        invtoken = Token(-1, '', filename, lineno)
                        raise globals.FSLParseError, (invtoken, 'Invalid indentation')
                    indent_steps = abs(cur_indent - prev_indent) // indent_length
                    if cur_indent > prev_indent:
                        for i in range(indent_steps):
                            indent_count += 1
                            yield Token(T_INDENT, '<INDENT>', filename, lineno)
                    elif cur_indent < prev_indent:
                        for i in range(indent_steps):
                            dedent_count += 1
                            yield Token(T_DEDENT, '<DEDENT>', filename, lineno)
                line = strippedline
                prev_indent = cur_indent

        if eof: break

        for pattern, ttype in _TOKENS_RE:
            match = re.match(pattern, line, re.UNICODE)
            if match:
                if ttype == _T_COMMENT:
                    line = ''
                    break

                tokenstr = match.group()
                line = line[len(tokenstr):].strip(_WHITESPACE)

                if ttype == T_BARE_STRING:
                    tokenlower = tokenstr.lower()
                    if tokenlower in _RESERVED_WORDS:
                        ttype = _RESERVED_WORDS[tokenlower]
                elif ttype == T_STRING:
                    assert tokenstr[0] == '"' and tokenstr[-1] == '"'
                    tokenstr = tokenstr[1:-1]
                elif ttype == T_LEFT_PAREN:
                    open_paren_count += 1
                elif ttype == T_RIGHT_PAREN:
                    open_paren_count -= 1

                line_had_content = True
                yield Token(ttype, tokenstr, filename, lineno)
                break
        else:
            invtoken = Token(-1, line.strip(), filename, lineno)
            raise globals.FSLParseError(invtoken, 'Invalid token')

        if (not line) and line_had_content and open_paren_count == 0:
            yield Token(T_LINE_END, '<LINE END>', filename, lineno)

    if cur_indent > 0:
        indent_steps = cur_indent // indent_length
        for i in range(indent_steps):
            yield Token(T_DEDENT, '<DEDENT>', filename, lineno)
            dedent_count += 1

    assert indent_count == dedent_count

    yield Token(T_EOF, '<EOF>', filename, lineno)
