import unittest

from fsl import lexer
from fsl import globals
import tutils

class TestLexer(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def _check_token_types(self, tokens, expected_types):
        for token, exp_type in zip(tokens, expected_types):
            self.assertEqual(token.tokentype, exp_type,
                'Expected type %s for token "%s", got %s' % (
                    lexer.type2str(exp_type), token.strvalue, lexer.type2str(token.tokentype)))

    def test_invalid(self):
        files = tutils.get_lexer_files('invalid*')
        self.failIf(len(files) == 0, 'No "invalid*" files found')
        for filename in files:
            try:
                tokens = list(lexer.iter_tokens(filename))
                self.fail('Expected ParseError for file "%s"' % filename)
            except globals.FSLParseError:
                pass

    def test_all_valid(self):
        """Test that all token types are recognized correctly."""
        expected_types = (lexer.T_BARE_STRING, lexer.T_STRING,
            lexer.T_INT, lexer.T_FLOAT,
            lexer.T_LESS_THAN, lexer.T_LESS_THAN_EQUAL, lexer.T_MORE_THAN,
            lexer.T_MORE_THAN_EQUAL, lexer.T_EQUAL, lexer.T_NOT_EQUAL,
            lexer.T_COMMA, lexer.T_LEFT_PAREN, lexer.T_RIGHT_PAREN,
            lexer.T_AND, lexer.T_EACH, lexer.T_IF, lexer.T_IN,
            lexer.T_NONREC, lexer.T_NOT, lexer.T_OR, lexer.T_LINE_END,
            lexer.T_INDENT, lexer.T_BARE_STRING, lexer.T_LINE_END, lexer.T_DEDENT,
            lexer.T_EOF)

        tokens = list(lexer.iter_tokens(tutils.get_lexer_file('valid_alltokens')))
        self._check_token_types(tokens, expected_types)

    def test_comment(self):
        tokens = list(lexer.iter_tokens(tutils.get_lexer_file('valid_comment')))
        self.failIf(len(tokens) != 10, 'Expected 10 valid tokens, got %d' % len(tokens))
        self.assertEqual(tokens[0].strvalue, 'abc',
            'First token expected to be "abc", got "%s"' % tokens[0].strvalue)
        self.assertEqual(tokens[6].strvalue, 'xyz#',
            '7th token expected to be "xyz#", got "%s"' % tokens[6].strvalue)

    def test_line_ends(self):
        tokens = list(lexer.iter_tokens(tutils.get_lexer_file('valid_line_ends')))
        expected_types = [lexer.T_BARE_STRING, lexer.T_LINE_END] * 5 + [lexer.T_EOF]
        self._check_token_types(tokens, expected_types)

    def test_begin_indent(self):
        tokens = list(lexer.iter_tokens(tutils.get_lexer_file('valid_begin_indent')))
        expected_types = [lexer.T_INDENT, lexer.T_BARE_STRING, lexer.T_LINE_END, lexer.T_DEDENT, lexer.T_EOF]
        self._check_token_types(tokens, expected_types)

    def test_eof(self):
        files = tutils.get_all_files('valid*')
        self.failIf(len(files) == 0, 'No "valid*" files found')
        for filename in files:
            tokens = list(lexer.iter_tokens(filename))
            self.assertEqual(tokens[-1].tokentype, lexer.T_EOF,
                'No EOF token at the end of "%s"' % filename)

    def test_balanced_indent(self):
        files = tutils.get_all_files('valid*')
        self.failIf(len(files) == 0, 'No "valid*" files found')
        for filename in files:
            indent = 0
            dedent = 0
            for token in lexer.iter_tokens(filename):
                if token.tokentype == lexer.T_INDENT:
                    indent += 1
                elif token.tokentype == lexer.T_DEDENT:
                    dedent += 1
            self.assertEqual(indent, dedent,
                'Unbalanced indent-dedent tokens in "%s" (%d indents, %d dedents)' % (filename, indent, dedent))

    def test_indent(self):
        tokens = list(lexer.iter_tokens(tutils.get_lexer_file('valid_indent')))
        indent = [t for t in tokens if t.tokentype == lexer.T_INDENT]
        dedent = [t for t in tokens if t.tokentype == lexer.T_DEDENT]

        self.assertEqual(len(indent), len(dedent),
            'Unbalanced indent-dedent tokens (%d indents, %d dedents)' % (len(indent), len(dedent)) )

        self.assertEqual(len(indent), 9,
            'Expected 9 indent tokens, got %d' % len(indent) )

    def test_empty(self):
        tokens = list(lexer.iter_tokens(tutils.get_lexer_file('valid_empty')))
        self.failUnless(len(tokens) == 1 and tokens[0].tokentype == lexer.T_EOF)

    def test_line_numbers(self):
        tokens = list(lexer.iter_tokens(tutils.get_lexer_file('valid_line_numbers')))
        expected_linenos = [3, 3, 4, 4, 6, 6, 9, 9, 9]
        for token, exp_lineno in zip(tokens, expected_linenos):
            self.assertEquals(token.lineno, exp_lineno,
                'Expected line number %d for token "%s", got %d' % (exp_lineno, token.strvalue, token.lineno))

    def test_whitespace_end(self):
        tokens = list(lexer.iter_tokens(tutils.get_lexer_file('valid_whitespace_end')))
        expected_types = [lexer.T_INT, lexer.T_LINE_END, lexer.T_EOF]
        self._check_token_types(tokens, expected_types)

    def test_large_numbers(self):
        tokens = list(lexer.iter_tokens(tutils.get_lexer_file('valid_large_numbers')))
        expected_types = [lexer.T_INT, lexer.T_FLOAT, lexer.T_LINE_END, lexer.T_EOF]
        self._check_token_types(tokens, expected_types)

    ### Multi-line expressions #########################################

    def _generic_multiline_expr(self, number):
        src_filename = tutils.get_lexer_file('valid_multilineexpr_%d' % number)
        tokens = list(lexer.iter_tokens(src_filename))
        expected_types = [lexer.T_LEFT_PAREN,
            lexer.T_INT, lexer.T_INT, lexer.T_INT, lexer.T_RIGHT_PAREN,
            lexer.T_LINE_END, lexer.T_EOF]
        self._check_token_types(tokens, expected_types)

    def test_multiline_expr_1(self):
        self._generic_multiline_expr(1)

    def test_multiline_expr_2(self):
        self._generic_multiline_expr(2)

def main(args):
    unittest.main()

if __name__=='__main__':
    import sys
    main(sys.argv[1:])
